/*
 * si4708icx.c
 *
 * Si4708 tuner on icx driver
 *
 * Copyright (c) 2011-2013,2015,2018 Sony Corporation
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/************/
/*@ headers */
/************/

#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/pm.h>
#include <linux/delay.h>
#include <linux/i2c.h>
#include <linux/videodev2.h>
#include <linux/switch.h>
#include <linux/proc_fs.h>

#include <media/v4l2-device.h>
#include <media/v4l2-ioctl.h>

#ifdef CONFIG_REGMON_DEBUG
#include <mach/regmon.h>
#endif

/***************/
/*@ parameters */
/***************/

/* license */
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Sony Corporation");
MODULE_DESCRIPTION("device driver for si4708 fm tuner");

/* module name */
#define MODULE_NAME SI4708ICX_RESOURCE_NAME

/* debug switch */
/* #define TRACE_PRINT_ON */
/* #define VALUE_PRINT_ON */
/* #define ELECTROSTATIC_DAMAGE_REPAIRER */
#define RSSI_RAW_VALUE_CONTRL
#define RESTORE_MUTE_STATE
#define UNDEF_SWITCH_DEV
#define USE_DMA
#define USE_MTK_GPIO_CONTROL
#ifdef USE_MTK_GPIO_CONTROL
#include <mach/mt_gpio.h>
#else
#include <linux/gpio.h>
#endif /* USE_MTK_GPIO_CONTROL */

/* register initial value */
#define SI4708_REG_SYSCONFIG1_INIT_VAL 0x00AA /* BLNDADJ=-12dB,GPO=low                       */
#define SI4708_REG_SYSCONFIG2_INIT_VAL 0x126F /* SEEKTH=0x12,BAND=JP-Wide,SPACE=50KHz,VOL=15 */
#define SI4708_REG_SYSCONFIG3_INIT_VAL 0x1000 /* SMUTEA=14dB                                 */

/* timing */
#define SI4708_SEEK_TUNE_TIME          60 /* from spec */
#define SI4708_TUNE_COMP_TIME          20
#define SI4708_TUNE_STOP_TIME           2
#define SI4708_STC_POLL_INTERVAL        5
#define SI4708_RESET_LOW_TIME           1
#define SI4708_ENABLE_COMP_TIME       111
#define SI4708_AUDIO_MODE_CHANGE_TIME 200

/* tune polling count */
#define SI4708_TUNE_POLLING_COUNT ((SI4708_SEEK_TUNE_TIME-SI4708_TUNE_COMP_TIME)/SI4708_STC_POLL_INTERVAL)

/****************/
/*@ definitions */
/****************/
#include<linux/platform_data/icx_si4708.h>
#include "si4708icx.h"

/* generic */
#ifndef FALSE
#define FALSE 0
#define TRUE  1
#endif
#define OFF   0
#define ON	  1
#define LOW   0
#define HIGH  1

/* de-emphasis mode */
#define DEEMPHASIS_MODE_75US 0
#define DEEMPHASIS_MODE_50US 1

/* sense mode */
#define SENSE_MODE_LOW  0
#define SENSE_MODE_HIGH 1

/* audio mode */
#define AUDIO_MODE_STEREO 0
#define AUDIO_MODE_MONO   1

/* power mode */
#define POWER_MODE_STANBY 0
#define POWER_MODE_ACTIVE 1

/* si4708 register */
#define SI4708_REG_DEVICEID    0x00 /* R   */
#define SI4708_REG_CHIPID      0x01 /* R   */
#define SI4708_REG_POWERCFG    0x02 /* R/W */
#define SI4708_REG_CHANNEL     0x03 /* R/W */
#define SI4708_REG_SYSCONFIG1  0x04 /* R/W */
#define SI4708_REG_SYSCONFIG2  0x05 /* R/W */
#define SI4708_REG_SYSCONFIG3  0x06 /* R/W */
#define SI4708_REG_TEST1       0x07 /* R/W */
#define SI4708_REG_TEST2       0x08 /* R/W */
#define SI4708_REG_BOOTCONFIG  0x09 /* R/W */
#define SI4708_REG_STATUS_RSSI 0x0A /* R   */
#define SI4708_REG_READCHAN    0x0B /* R   */
#define SI4708_REG_RDSA        0x0C /* R   */
#define SI4708_REG_RDSB        0x0D /* R   */
#define SI4708_REG_RDSC        0x0E /* R   */
#define SI4708_REG_RDSD        0x0F /* R   */

/* power register bit assign */
#define SI4708_REG_POWERCFG_DMUTE         0x4000
#define SI4708_REG_POWERCFG_MONO          0x2000
#define SI4708_REG_POWERCFG_DISABLE       0x0040
#define SI4708_REG_POWERCFG_ENABLE        0x0001

/* channel register bit assign */
#define SI4708_REG_CHANNEL_CHANNLE_MASK   0x03FF
#define SI4708_REG_CHANNEL_TUNE           0x8000

/* config 1 register bit assign */
#define SI4708_REG_SYSCONFIG1_DE          0x0800

/* config 2 register bit assign */
#define SI4708_REG_SYSCONFIG2_SEEKTH_MASK 0xFF00

/* status register bit assign */
#define SI4708_REG_STATUS_RSSI_STC        0x4000
#define SI4708_REG_STATUS_RSSI_AFCRL      0x1000
#define SI4708_REG_STATUS_RSSI_ST         0x0100
#define SI4708_REG_STATUS_RSSI_RSSI_MASK  0x00FF

/* register access parameter */
#define SI4708_REG_COUNT   16
#define SI4708_REG_START    SI4708_REG_STATUS_RSSI
#define SI4708_RREG_OFFSET  0
#define SI4708_WREG_OFFSET  8

/* register access count */
#define SI4708_RREG_COUNT 16
#define SI4708_WREG_COUNT  8

/* debug print */
#ifdef TRACE_PRINT_ON
	#define print_trace(fmt, args...) printk(KERN_INFO "------- " fmt, ##args)
#else
	#define print_trace(fmt, args...)
#endif
#ifdef VALUE_PRINT_ON
	#define print_value(fmt, args...) printk(KERN_INFO "        " fmt, ##args)
#else
	#define print_value(fmt, args...)
#endif

/* general */
#define minimum(_a,_b) ( (_a)<(_b) ? (_a) : (_b) )
#define maximum(_a,_b) ( (_a)>(_b) ? (_a) : (_b) )

/* model parameters */
typedef struct {
	int           deemphasis;
	unsigned char sense_high;
	unsigned char sense_low;
	int           minimum_frequency;
	int           frequency_step;
} MODEL_PARAMETERS;

/* private data */
typedef struct {
	struct v4l2_device    v4l2_device;
	struct video_device * video_device;
	struct i2c_client   * i2c_client;
	unsigned short        reg_cache[SI4708_REG_COUNT];
	struct mutex          global_mutex;
	int                   user_count;
	int                   tuned_freq;
	int                   demph_mode;
	int                   sense_mode;
	int                   audio_mode;
	int                   mute_state;
	int                   power_mode;
	int                   sus_pow_mode;
} UNIT_PARAMETERS;

/***************/
/*@ prototypes */
/***************/

/* entry_routins */
static int  __init    si4708icx_init(void);
static void __exit    si4708icx_exit(void);
static int  si4708icx_probe(
	      struct i2c_client    * i2c,
	const struct i2c_device_id * identification
);
static int  si4708icx_remove(
	struct i2c_client * i2c
);
static int  si4708icx_suspend(
	struct i2c_client * i2c,
	pm_message_t        message
);
static int  si4708icx_resume(
	struct i2c_client * i2c
);
static void si4708icx_shutdown(
	struct i2c_client * i2c
);

/* operation_routines */
static int  si4708icx_open(
	struct file * file
);
static int  si4708icx_close(
	struct file * file
);
static int si4708icx_querycap(
	struct file            * file,
	void                   * private,
	struct v4l2_capability * capability
);
static int si4708icx_s_tuner(
	struct file       * file,
	void              * private,
	const struct v4l2_tuner * tuner
);
static int si4708icx_g_tuner(
	struct file       * file,
	void              * private,
	struct v4l2_tuner * tuner
);
static int si4708icx_s_frequency(
	struct file           * file,
	void                  * private,
	const struct v4l2_frequency * frequency
);
static int si4708icx_g_frequency(
	struct file           * file,
	void                  * private,
	struct v4l2_frequency * frequency
);
static int si4708icx_queryctrl(
	struct file           * file,
	void                  * private,
	struct v4l2_queryctrl * query
);
static int si4708icx_s_ctrl(
	struct file         * file,
	void                * private,
	struct v4l2_control * control
);
static int si4708icx_g_ctrl(
	struct file         * file,
	void                * private,
	struct v4l2_control * control
);

/* control_routines */
static int setup_platform (UNIT_PARAMETERS * unit);
static int reset_platform (UNIT_PARAMETERS * unit);
static int poweron_chip   (UNIT_PARAMETERS * unit);
static int poweroff_chip  (UNIT_PARAMETERS * unit);
static int enable_chip    (UNIT_PARAMETERS * unit);
static int disable_chip   (UNIT_PARAMETERS * unit);
static int reset_chip     (UNIT_PARAMETERS * unit);
static int tune           (UNIT_PARAMETERS * unit, int frequency);
static int set_audio_mode (UNIT_PARAMETERS * unit, int value);
static int set_sense_mode (UNIT_PARAMETERS * unit, int value);
static int switch_mute    (UNIT_PARAMETERS * unit, int value);
#ifdef ELECTROSTATIC_DAMAGE_REPAIRER
static int fix_output     (void);
#endif /* ELECTROSTATIC_DAMAGE_REPAIRER */
/* basic_control_routines */
static int bcr_set_frequency         (UNIT_PARAMETERS * unit, int value);
static int bcr_set_tune              (UNIT_PARAMETERS * unit, int value);
static int bcr_wait_tune_complete    (UNIT_PARAMETERS * unit);
static int bcr_check_signal_strength (UNIT_PARAMETERS * unit);
#ifdef RSSI_RAW_VALUE_CONTRL
static int bcr_get_signal_strength   (UNIT_PARAMETERS * unit, int * value);
#endif /* RSSI_RAW_VALUE_CONTRL */
static int bcr_set_deemphasis_mode   (UNIT_PARAMETERS * unit, int value);
static int bcr_set_sense_mode        (UNIT_PARAMETERS * unit, int value);
static int bcr_set_audio_mode        (UNIT_PARAMETERS * unit, int value);
static int bcr_switch_mute           (UNIT_PARAMETERS * unit, int value);
static int bcr_set_power_mode        (UNIT_PARAMETERS * unit, int value);
static int bcr_reset_chip            (UNIT_PARAMETERS * unit);

/* register_access_routines */
static int rar_modify_reg(
	UNIT_PARAMETERS * unit,
	unsigned char     address,
	unsigned short    value,
	unsigned short    mask
);
static int rar_write_reg(
	UNIT_PARAMETERS * unit,
	unsigned char     address,
	unsigned short    value
);
static int rar_read_reg(
	UNIT_PARAMETERS * unit,
	unsigned char     address,
	unsigned short  * value
);
static int rar_write_all_reg(
	UNIT_PARAMETERS * unit
);
static int rar_read_all_reg(
	UNIT_PARAMETERS * unit
);
#ifdef ELECTROSTATIC_DAMAGE_REPAIRER
extern void cxd3774gf_register_electrostatic_damage_repairer(int (*function)(void));
#endif
#ifdef CONFIG_REGMON_DEBUG
/* regmon_routines */
static int si4708icx_regmon_add(void * data);
static int si4708icx_regmon_del(void);
#endif

/**************/
/*@ variables */
/**************/

module_init(si4708icx_init);
module_exit(si4708icx_exit);

static const struct i2c_device_id si4708icx_i2c_id[] = {
	{SI4708ICX_RESOURCE_NAME, 0},
	{}
};
MODULE_DEVICE_TABLE(i2c, si4708icx_i2c_id);

static struct i2c_driver si4708icx_i2c_driver = {
	.driver   = {
		.name  = MODULE_NAME,
		.owner = THIS_MODULE,
	},
	.id_table = si4708icx_i2c_id,
	.probe    = si4708icx_probe,
	.remove   = si4708icx_remove,
	.suspend  = si4708icx_suspend,
	.resume   = si4708icx_resume,
	.shutdown = si4708icx_shutdown,
};

static const struct v4l2_file_operations si4708icx_file_operations = {
	.owner           = THIS_MODULE,
	.open            = si4708icx_open,
	.release         = si4708icx_close,
	.unlocked_ioctl  = video_ioctl2,
};

static const struct v4l2_ioctl_ops si4708icx_ioctl_operations = {
	.vidioc_querycap       = si4708icx_querycap,
	.vidioc_s_tuner        = si4708icx_s_tuner,
	.vidioc_g_tuner        = si4708icx_g_tuner,
	.vidioc_s_frequency    = si4708icx_s_frequency,
	.vidioc_g_frequency    = si4708icx_g_frequency,
	.vidioc_queryctrl      = si4708icx_queryctrl,
	.vidioc_s_ctrl         = si4708icx_s_ctrl,
	.vidioc_g_ctrl         = si4708icx_g_ctrl,
};

static MODEL_PARAMETERS model_parameters = {
	.deemphasis        = DEEMPHASIS_MODE_50US,
	.sense_high        = 0x12,
	.sense_low         = 0x2B,
	.minimum_frequency = 76000,
	.frequency_step    = 50,
};

static UNIT_PARAMETERS *global_unit = NULL;

static int DeEmphasis=50;
module_param_named(deemphasis,DeEmphasis,int,0);
MODULE_PARM_DESC(deemphasis,"deemphasis.");
#ifndef UNDEF_SWITCH_DEV
#ifdef CONFIG_SWITCH

static struct switch_dev fm_tuner_switch_dev = {
	.name = "fm_tuner",
};

#endif
#endif
/*******************/
/*@ entry_routines */
/*******************/

static int __init si4708icx_init(void)
{
	int rv;

	printk(KERN_INFO "%s()\n",__FUNCTION__);

	rv=i2c_add_driver(&si4708icx_i2c_driver);
	if(rv<0) {
		printk(KERN_ERR "%s(): i2c_add_driver() error\n",__FUNCTION__);
		return(rv);
	}

	return(0);
}

static void __exit si4708icx_exit(void)
{
	print_trace("%s()\n",__FUNCTION__);

	i2c_del_driver(&si4708icx_i2c_driver);

	return;
}

static int si4708icx_probe(
		  struct i2c_client    * i2c,
	const struct i2c_device_id * identification
)
{
	UNIT_PARAMETERS * unit;
	int rv;

	printk(KERN_INFO "%s(): deemphasis mode = %s\n",__FUNCTION__, (DeEmphasis==750 ? "75us" : "50us"));

	print_value("deemphasis parameter = %d\n",DeEmphasis);

	if(DeEmphasis==750)
		model_parameters.deemphasis=DEEMPHASIS_MODE_75US;
	else
		model_parameters.deemphasis=DEEMPHASIS_MODE_50US;

	unit=kzalloc(sizeof(UNIT_PARAMETERS),GFP_KERNEL);
	if(unit==NULL){
		printk(KERN_ERR "%s(): kzalloc() error\n",__FUNCTION__);
		return(-ENOMEM);
	}

	mutex_init(&unit->global_mutex);
	unit->i2c_client   = i2c;
	unit->user_count   = 0;
	unit->tuned_freq   = -1;
	unit->demph_mode   = model_parameters.deemphasis;
	unit->sense_mode   = SENSE_MODE_HIGH;
	unit->audio_mode   = AUDIO_MODE_MONO;
	unit->mute_state   = ON;
	unit->power_mode   = POWER_MODE_STANBY;
	unit->sus_pow_mode = -1;

	/* set private data to i2c structure */
	i2c_set_clientdata(i2c,unit);

	/* hardware initialize */

	rv=setup_platform(unit);
	if(rv<0){
		printk(KERN_ERR "%s(): setup_platform() error\n",__FUNCTION__);
		kfree(unit);
		return(rv);
	}

	rv=poweron_chip(unit);
	if(rv<0){
		printk(KERN_ERR "%s(): poweron_chip() error\n",__FUNCTION__);
		reset_platform(unit);
		kfree(unit);
		return(rv);
	}

	rv=reset_chip(unit);
	if(rv<0){
		printk(KERN_ERR "%s(): reset_chip() error\n",__FUNCTION__);
		poweroff_chip(unit);
		reset_platform(unit);
		kfree(unit);
		return(rv);
	}

	/* v4l2 initialize */

	unit->video_device=video_device_alloc();
	if(unit->video_device==NULL){
		printk(KERN_ERR "%s(): video_device_alloc() error\n",__FUNCTION__);
		poweroff_chip(unit);
		reset_platform(unit);
		kfree(unit);
		return(-ENOMEM);
	}

	strncpy(unit->video_device->name,"Silicon Labs. FM Tuner",sizeof(unit->video_device->name));
	/* unit->video_device->dev       =  i2c->dev; */
	unit->video_device->release   =  video_device_release_empty;
	unit->video_device->fops      = &si4708icx_file_operations;
	unit->video_device->ioctl_ops = &si4708icx_ioctl_operations;

	/* set private data to video structure */
	video_set_drvdata(unit->video_device,unit);

	/* rv=v4l2_device_register(&i2c->dev,&unit->v4l2_device); */
	/* if(rv<0){ */
		/* printk(KERN_ERR "%s(): v4l2_device_register() error\n",__FUNCTION__); */
		/* video_device_release(unit->video_device); */
		/* kfree(unit); */
		/* return(rv); */
	/* } */

	rv=video_register_device(unit->video_device,VFL_TYPE_RADIO,-1);
	if(rv<0) {
		printk(KERN_ERR "%s(): video_register_device() error\n",__FUNCTION__);
		/* v4l2_device_unregister(&unit->v4l2_device); */
		video_device_release(unit->video_device);
		poweroff_chip(unit);
		reset_platform(unit);
		kfree(unit);
		return(rv);
	}

#ifndef UNDEF_SWITCH_DEV
	#ifdef CONFIG_SWITCH
	rv=switch_dev_register(&fm_tuner_switch_dev);
	if(rv<0){
		printk(KERN_ERR "%s(): switch_dev_register() error\n",__FUNCTION__);
		video_unregister_device(unit->video_device);
		/* v4l2_device_unregister(&unit->v4l2_device); */
		video_device_release(unit->video_device);
		poweroff_chip(unit);
		reset_platform(unit);
		kfree(unit);
		return(rv);
	}
	#endif
#endif
	/* registering function recovering static electricity. */
	global_unit = unit;
#ifdef ELECTROSTATIC_DAMAGE_REPAIRER
	cxd3774gf_register_electrostatic_damage_repairer(fix_output);
#endif
	#ifdef CONFIG_REGMON_DEBUG
	si4708icx_regmon_add(unit);
	#endif

	return(0);
}

static int si4708icx_remove(
	struct i2c_client * i2c
)
{
	UNIT_PARAMETERS * unit;

	print_trace("%s()\n",__FUNCTION__);

	unit=i2c_get_clientdata(i2c);

#ifndef UNDEF_SWITCH_DEV
	#ifdef CONFIG_SWITCH
	switch_set_state(&fm_tuner_switch_dev,0);
	#endif
#endif /* UNDEF_SWITCH_DEV */
	#ifdef CONFIG_REGMON_DEBUG
	si4708icx_regmon_del();
	#endif
#ifdef ELECTROSTATIC_DAMAGE_REPAIRER
	/* unregistering function recovering static electricity. */
	cxd3774gf_register_electrostatic_damage_repairer(NULL);
#endif
	global_unit = NULL;
#ifndef UNDEF_SWITCH_DEV
	#ifdef CONFIG_SWITCH
	switch_dev_unregister(&fm_tuner_switch_dev);
	#endif
#endif /* UNDEF_SWITCH_DEV */

	video_unregister_device(unit->video_device);

	/* v4l2_device_unregister(&unit->v4l2_device); */

	video_device_release(unit->video_device);

	disable_chip(unit);

	reset_chip(unit);

	poweroff_chip(unit);

	reset_platform(unit);

	kfree(unit);

	return(0);
}

static int si4708icx_suspend(struct i2c_client * i2c, pm_message_t message)
{
	UNIT_PARAMETERS * unit;

	print_trace("%s()\n",__FUNCTION__);

	unit=i2c_get_clientdata(i2c);

	unit->sus_pow_mode=unit->power_mode;

	if(unit->power_mode==POWER_MODE_ACTIVE)
		disable_chip(unit);

	/* for gpio value not are fixed after gpio reset. */
	/* not needed for omap4430. */
	/* read value may collapse by I2C controller clock change on omap4430 dpll cascading. */
	/* confirm I2C clock setting. */
	rar_read_all_reg(unit);

	return(0);
}
static int si4708icx_resume(struct i2c_client * i2c)
{
	UNIT_PARAMETERS * unit;

	print_trace("%s()\n",__FUNCTION__);

	unit=i2c_get_clientdata(i2c);

	/* for gpio value are not fixed after gpio reset. */
	/* not needed for omap4430. */
	rar_write_all_reg(unit);

	if(unit->sus_pow_mode==POWER_MODE_ACTIVE)
		enable_chip(unit);

	return(0);
}

static void si4708icx_shutdown(struct i2c_client * i2c)
{
	UNIT_PARAMETERS * unit;

	print_trace("%s()\n",__FUNCTION__);

	unit=i2c_get_clientdata(i2c);
#ifdef ELECTROSTATIC_DAMAGE_REPAIRER
	/* unregistering function recovering static electricity. */
	cxd3774gf_register_electrostatic_damage_repairer(NULL);
#endif
	global_unit = NULL;

	disable_chip(unit);

	reset_chip(unit);

	poweroff_chip(unit);

	reset_platform(unit);

	return;
}

/***********************/
/*@ operation_routines */
/***********************/

static int si4708icx_open(struct file * file)
{
	UNIT_PARAMETERS  * unit;

	printk(KERN_INFO "%s()\n",__FUNCTION__);

	unit=video_drvdata(file);

	mutex_lock(&unit->global_mutex);

	unit->user_count++;

	if(unit->user_count==1){
#ifndef UNDEF_SWITCH_DEV
		#ifdef CONFIG_SWITCH
		switch_set_state(&fm_tuner_switch_dev,1);
		#endif
#endif
		enable_chip(unit);
	}

	mutex_unlock(&unit->global_mutex);

	return(0);
}

static int si4708icx_close(struct file * file)
{
	UNIT_PARAMETERS * unit;

	printk(KERN_INFO "%s()\n",__FUNCTION__);

	unit=video_drvdata(file);

	mutex_lock(&unit->global_mutex);

	if(unit->user_count > 0)
		unit->user_count--;

	if(unit->user_count==0){
		disable_chip(unit);
#ifndef UNDEF_SWITCH_DEV
		#ifdef CONFIG_SWITCH
		switch_set_state(&fm_tuner_switch_dev,0);
		#endif
#endif
	}

	mutex_unlock(&unit->global_mutex);

	return(0);
}

static int si4708icx_querycap(
	struct file            * file,
	void                   * dummy,
	struct v4l2_capability * capability
)
{
	print_trace("%s()\n",__FUNCTION__);

	strlcpy(capability->driver,   "si4708 icx dirver",    sizeof(capability->driver));
	strlcpy(capability->card,     SI4708ICX_RESOURCE_NAME,sizeof(capability->card));
	strlcpy(capability->bus_info, "i2c-2",                sizeof(capability->card));
	capability->version      = KERNEL_VERSION(1, 0, 0);
	capability->capabilities = V4L2_CAP_TUNER | V4L2_CAP_RADIO;

	return(0);
}

static int si4708icx_s_tuner(
	struct file       * file,
	void              * dummy,
	const struct v4l2_tuner * tuner
)
{
	UNIT_PARAMETERS * unit;
	int val;
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	unit=video_drvdata(file);

	if(tuner->index!=0){
		printk(KERN_ERR "%s(): invalid tuner %d\n",__FUNCTION__,tuner->index);
		return(-EINVAL);
	}

	if(tuner->audmode==V4L2_TUNER_MODE_STEREO){
		val=AUDIO_MODE_STEREO;
	}
	else if(tuner->audmode==V4L2_TUNER_MODE_MONO){
		val=AUDIO_MODE_MONO;
	}
	else{
		printk(KERN_ERR "%s(): invalid audio mode %d\n",__FUNCTION__,tuner->audmode);
		return(-EINVAL);
	}

	rv=set_audio_mode(unit,val);
	if(rv<0){
		printk(KERN_ERR "%s(): set_audio_mode() error\n",__FUNCTION__);
		return(rv);
	}

	return(0);
}

static int si4708icx_g_tuner(
	struct file       * file,
	void              * dummy,
	struct v4l2_tuner * tuner
)
{
	UNIT_PARAMETERS * unit;
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	unit=video_drvdata(file);

	if(tuner->index!=0){
		printk(KERN_ERR "%s(): invalid tuner %d\n",__FUNCTION__,tuner->index);
		return(-EINVAL);
	}

	strlcpy(tuner->name,"FM",sizeof(tuner->name));

	tuner->type = V4L2_TUNER_RADIO;

	tuner->capability = V4L2_TUNER_CAP_LOW | V4L2_TUNER_CAP_STEREO;

	tuner->rangelow  =  76 * 16000; /*  76MHz =  76 * 1000 * 1000 / 62.5 */
	tuner->rangehigh = 108 * 16000; /* 108MHz = 108 * 1000 * 1000 / 62.5 */

	/* rv=get_stereo_status(unit); */
	/* if(rv<0){ */
		/* printk(KERN_ERR "%s(): get_stereo_status() error\n",__FUNCTION__); */
		/* return(rv); */
	/* } */
	/* if(rv) */
		/* tuner->rxsubchans = V4L2_TUNER_SUB_STEREO | V4L2_TUNER_SUB_MONO; */
	/* else */
		/* tuner->rxsubchans = V4L2_TUNER_SUB_MONO; */
	tuner->rxsubchans = V4L2_TUNER_SUB_STEREO | V4L2_TUNER_SUB_MONO;

	if(unit->audio_mode==AUDIO_MODE_STEREO)
		tuner->audmode = V4L2_TUNER_MODE_STEREO;
	else
		tuner->audmode = V4L2_TUNER_MODE_MONO;

	rv=bcr_check_signal_strength(unit);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_check_signal_strength() error\n",__FUNCTION__);
		return(rv);
	}
	tuner->signal = 0xFFFF * rv;

	/* not support */
	tuner->afc = 0;

	return(0);
}

static int si4708icx_s_frequency(
	struct file           * file,
	void                  * dummy,
	const struct v4l2_frequency * frequency
)
{
	UNIT_PARAMETERS * unit;
	unsigned int freq;
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	unit=video_drvdata(file);

	if(frequency->tuner!=0){
		printk(KERN_ERR "%s(): invalid tuner %d\n",__FUNCTION__,frequency->tuner);
		return(-EINVAL);
	}

	freq=frequency->frequency>>4;

	if( (freq<76000) || (freq>108000) ){
		printk(KERN_ERR "%s(): invalid frequency %d\n",__FUNCTION__,(int)freq);
		return(EINVAL);
	}

	rv=tune(unit,(int)freq);
	if(rv<0){
		printk(KERN_ERR "%s(): tune() error\n",__FUNCTION__);
		return(rv);
	}

	return(0);
}

static int si4708icx_g_frequency(
	struct file           * file,
	void                  * dummy,
	struct v4l2_frequency * frequency
)
{
	UNIT_PARAMETERS * unit;

	print_trace("%s()\n",__FUNCTION__);

	unit=video_drvdata(file);

	if(frequency->tuner!=0){
		printk(KERN_ERR "%s(): invalid tuner %d\n",__FUNCTION__,frequency->tuner);
		return(-EINVAL);
	}

	frequency->type = V4L2_TUNER_RADIO;
	frequency->frequency=(unsigned int)unit->tuned_freq * 16;

	return(0);
}

static int si4708icx_queryctrl(
	struct file           * file,
	void                  * dummy,
	struct v4l2_queryctrl * query
)
{
	int ret;

	print_trace("%s()\n",__FUNCTION__);

	switch(query->id){
		case V4L2_CID_AUDIO_MUTE:
			ret=v4l2_ctrl_query_fill(query,0,1,1,0);
			break;
		case V4L2_CID_SI4708ICX_SENSE_MODE:
			ret=v4l2_ctrl_query_fill(query,0,0x7F,1,0x06);
			break;
		default:
			return(-EINVAL);
	}

	return(ret);
}

static int si4708icx_s_ctrl(
	struct file         * file,
	void                * dummy,
	struct v4l2_control * control
)
{
	UNIT_PARAMETERS * unit;

	print_trace("%s()\n",__FUNCTION__);

	unit=video_drvdata(file);

	switch(control->id){

		case V4L2_CID_AUDIO_MUTE:
			if(control->value)
				switch_mute(unit,ON);
			else
				switch_mute(unit,OFF);
			break;

		case V4L2_CID_SI4708ICX_SENSE_MODE:
			set_sense_mode(unit,control->value);
			break;

		case V4L2_CID_SI4708ICX_SENSE_THRESHOLD_HIGH:
			model_parameters.sense_high=(unsigned char)(control->value&0xFF);
			bcr_set_sense_mode(unit,unit->sense_mode);
			break;

		case V4L2_CID_SI4708ICX_SENSE_THRESHOLD_LOW:
			model_parameters.sense_low=(unsigned char)(control->value&0xFF);
			bcr_set_sense_mode(unit,unit->sense_mode);
			break;

		/* for test tool */
		case V4L2_CID_SI4708ICX_USER_COUNT:
			mutex_lock(&unit->global_mutex);

			if(unit->user_count + control->value < 1){
				unit->user_count = 1;
			}else{
				unit->user_count += control->value;
			}

			mutex_unlock(&unit->global_mutex);
			break;

		default:
			return(-EINVAL);
	}

	return(0);
}

static int si4708icx_g_ctrl(
	struct file         * file,
	void                * dummy,
	struct v4l2_control * control
)
{
	UNIT_PARAMETERS * unit;
#ifdef RSSI_RAW_VALUE_CONTRL
	int rssi_raw = 0;
	int rv = 0;
#endif /* RSSI_RAW_VALUE_CONTRL */

	print_trace("%s()\n",__FUNCTION__);

	unit=video_drvdata(file);

	switch(control->id){

		case V4L2_CID_AUDIO_MUTE:
			control->value=unit->mute_state;
			break;

		case V4L2_CID_SI4708ICX_SENSE_MODE:
			control->value=unit->sense_mode;
			break;

		case V4L2_CID_SI4708ICX_SENSE_THRESHOLD_HIGH:
			control->value=model_parameters.sense_high;
			break;

		case V4L2_CID_SI4708ICX_SENSE_THRESHOLD_LOW:
			control->value=model_parameters.sense_low;
			break;

		/* for test tool */
		case V4L2_CID_SI4708ICX_USER_COUNT:
			control->value=unit->user_count;
			break;
#ifdef RSSI_RAW_VALUE_CONTRL
		case V4L2_CID_SI4708ICX_RAW_SIGNAL_STRENGTH:
            rv = bcr_get_signal_strength(unit, &rssi_raw);
	        if(rv<0){
		        printk(KERN_ERR "%s(): bcr_get_signal_strength() error\n",__FUNCTION__);
		        return(rv);
	        }
			control->value=rssi_raw;
			break;
#endif /* RSSI_RAW_VALUE_CONTRL */
		default:
			return(-EINVAL);
	}

	return(0);
}

/*********************/
/*@ control_routines */
/*********************/

static int setup_platform(UNIT_PARAMETERS * unit)
{
	SI4708ICX_PLATFORM_RESOURCE * resource;
	int rv;

	print_trace("%s()\n",__FUNCTION__);

    /* SETUP GPIO >>> */
	mt_set_gpio_dir(GPIO278 ,GPIO_DIR_IN);
    mt_set_gpio_mode(GPIO278,GPIO_MODE_GPIO);
    mt_set_gpio_pull_select(GPIO278,GPIO_PULL_UP);
    mt_set_gpio_pull_enable(GPIO278,GPIO_PULL_ENABLE);
    /* <<< SETUP GPIO */

	resource = unit->i2c_client->dev.platform_data;
#ifndef USE_MTK_GPIO_CONTROL
    rv=gpio_request(resource->port_reset,"FM_RST");
	if(rv<0){
		printk(KERN_ERR "gpio_request() error\n");
		return(rv);
	}
#endif /* !USE_MTK_GPIO_CONTROL */
#ifdef USE_MTK_GPIO_CONTROL
    rv=mt_set_gpio_out(resource->port_reset,1);
#else
	rv=gpio_direction_output(resource->port_reset,1);
#endif /* USE_MTK_GPIO_CONTROL */
	if(rv<0){
		printk(KERN_ERR "gpio_direction_output() error\n");
#ifndef USE_MTK_GPIO_CONTROL
		gpio_free(resource->port_reset);
#endif /* !USE_MTK_GPIO_CONTROL */
		return(rv);
	}

	return(0);
}

static int reset_platform(UNIT_PARAMETERS * unit)
{
	SI4708ICX_PLATFORM_RESOURCE * resource;

	print_trace("%s()\n",__FUNCTION__);

	resource = unit->i2c_client->dev.platform_data;

	/* keep gpio settings for chip bug. */

	/* gpio_direction_input(resource->port_reset); */
	/* gpio_free(resource->port_reset); */

	return(0);
}

static int poweron_chip(UNIT_PARAMETERS * unit)
{
	print_trace("%s()\n",__FUNCTION__);

	/* no action, always power on. */

	return(0);
}

static int poweroff_chip(UNIT_PARAMETERS * unit)
{
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	/* keep power on to keep gpio settings. */
	/* use standby mode.                    */

	rv=bcr_set_power_mode(unit,POWER_MODE_STANBY);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_set_power_mode() error\n",__FUNCTION__);
		return(rv);
	}

	unit->power_mode=POWER_MODE_STANBY;

	return(0);
}

static int enable_chip(UNIT_PARAMETERS * unit)
{
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	rv=bcr_set_power_mode(unit,POWER_MODE_ACTIVE);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_set_power_mode() error\n",__FUNCTION__);
		return(rv);
	}

	unit->power_mode=POWER_MODE_ACTIVE;

	return(0);
}

static int disable_chip(UNIT_PARAMETERS * unit)
{
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	rv=bcr_set_power_mode(unit,POWER_MODE_STANBY);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_set_power_mode() error\n",__FUNCTION__);
		return(rv);
	}

	unit->power_mode=POWER_MODE_STANBY;

	return(0);
}

static int reset_chip(UNIT_PARAMETERS * unit)
{
	unsigned short chip_id;
	unsigned short dev_id;
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	/* reset */
	rv=bcr_reset_chip(unit);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_reset_chip() error\n",__FUNCTION__);
		return(rv);
	}

	/* read all reg to chache */
	rv=rar_read_all_reg(unit);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_read_all_reg() error\n",__FUNCTION__);
		return(rv);
	}

	/* config 1 */
	rv=rar_write_reg(unit,SI4708_REG_SYSCONFIG1,SI4708_REG_SYSCONFIG1_INIT_VAL);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_write_reg() error\n",__FUNCTION__);
		return(rv);
	}

	/* config 2 */
	rv=rar_write_reg(unit,SI4708_REG_SYSCONFIG2,SI4708_REG_SYSCONFIG2_INIT_VAL);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_write_reg() error\n",__FUNCTION__);
		return(rv);
	}

	/* config 3 */
	rv=rar_write_reg(unit,SI4708_REG_SYSCONFIG3,SI4708_REG_SYSCONFIG3_INIT_VAL);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_write_reg() error\n",__FUNCTION__);
		return(rv);
	}

	/* read device id */
	rv=rar_read_reg(unit,SI4708_REG_DEVICEID,&dev_id);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_read_reg() error\n",__FUNCTION__);
		return(rv);
	}

	/* read chip id */
	rv=rar_read_reg(unit,SI4708_REG_CHIPID,&chip_id);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_read_reg() error\n",__FUNCTION__);
		return(rv);
	}

	printk(KERN_INFO "%s: Chip ID: %04X , Device ID: %04X \n",MODULE_NAME,(unsigned int)chip_id,(unsigned int)dev_id);

	bcr_switch_mute(unit,unit->mute_state);
	bcr_set_deemphasis_mode(unit,unit->demph_mode);
	bcr_set_audio_mode(unit,unit->audio_mode);
	bcr_set_sense_mode(unit,unit->sense_mode);

	if(unit->power_mode==POWER_MODE_ACTIVE)
		enable_chip(unit);

	return(0);
}

static int tune(UNIT_PARAMETERS * unit, int frequency)
{
	int result;
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	rv=bcr_switch_mute(unit,ON);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_switch_mute(ON) error\n",__FUNCTION__);
		return(rv);
	}

	rv=bcr_set_frequency(unit,frequency);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_set_frequency() error\n",__FUNCTION__);
		return(rv);
	}

	rv=bcr_set_tune(unit,ON);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_set_tune(ON) error\n",__FUNCTION__);
		return(rv);
	}

	rv=bcr_wait_tune_complete(unit);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_wait_tune_complete() error\n",__FUNCTION__);
		return(rv);
	}

	result=rv;

	rv=bcr_set_tune(unit,OFF);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_set_tune(OFF) error\n",__FUNCTION__);
		return(rv);
	}

#ifdef RESTORE_MUTE_STATE
	rv=bcr_switch_mute(unit,unit->mute_state);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_switch_mute(restore) error\n",__FUNCTION__);
		return(rv);
	}
#else
	rv=bcr_switch_mute(unit,OFF);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_switch_mute(OFF) error\n",__FUNCTION__);
		return(rv);
	}
	unit->mute_state=OFF;
#endif /* RESTORE_MUTE_STATE */
	unit->tuned_freq=frequency;

	return(result);
}

static int set_sense_mode(UNIT_PARAMETERS * unit, int value)
{
	int rv;

	print_trace("%s(%d)\n",__FUNCTION__,value);

	rv=bcr_set_sense_mode(unit,value);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_set_sense_mode() error\n",__FUNCTION__);
		return(rv);
	}

	unit->sense_mode=value;

	return(0);
}

static int set_audio_mode(UNIT_PARAMETERS * unit, int value)
{
	int rv;

	print_trace("%s(%d)\n",__FUNCTION__,value);

	rv=bcr_set_audio_mode(unit,value);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_set_audio_mode() error\n",__FUNCTION__);
		return(rv);
	}

	unit->audio_mode=value;

	return(0);
}

static int switch_mute(UNIT_PARAMETERS * unit, int value)
{
	int rv;

	print_trace("%s(%d)\n",__FUNCTION__,value);

	rv=bcr_switch_mute(unit,value);
	if(rv<0){
		printk(KERN_ERR "%s(): bcr_switch_mute() error\n",__FUNCTION__);
		return(rv);
	}

	unit->mute_state=value;

	return(0);
}

#ifdef ELECTROSTATIC_DAMAGE_REPAIRER
/* for static electricity */
static int fix_output(void)
{
	int rv;
	UNIT_PARAMETERS * unit;

	print_trace("%s()\n",__FUNCTION__);

	if(global_unit==NULL)
		return(-ENXIO);
	unit = global_unit;

	if(unit->user_count==0){
		printk(KERN_WARNING "%s(): %s is not opened.\n",__FUNCTION__, MODULE_NAME);
		return(0);
	}

	rv=switch_mute(unit, ON);
	if(rv<0){
		printk(KERN_ERR "%s(): switch_mute() error\n",__FUNCTION__);
		return(rv);
	}

	rv=reset_chip(unit);
	if(rv<0){
		printk(KERN_ERR "%s(): reset_chip() error\n",__FUNCTION__);
		return(rv);
	}

	rv=tune(unit, unit->tuned_freq);
	if(rv<0){
		printk(KERN_ERR "%s(): tune() error\n",__FUNCTION__);
		return(rv);
	}

	printk(KERN_WARNING "%s(): %s fixed by static electricity.\n",__FUNCTION__, MODULE_NAME);

	return(0);
}
#endif /* ELECTROSTATIC_DAMAGE_REPAIRER */
/***************************/
/*@ basic_control_routines */
/***************************/

static int bcr_set_frequency(UNIT_PARAMETERS * unit, int value)
{
	unsigned short us;
	int rv;

	printk(KERN_INFO "%s(%d)\n",__FUNCTION__,value);

	if( (value<76000) || (value>108000) ){
		printk(KERN_ERR "%s(): invalid frequency %d\n",__FUNCTION__,value);
		return(-EINVAL);
	}

	us = (value - model_parameters.minimum_frequency) / model_parameters.frequency_step;

	rv=rar_modify_reg(
		unit,
		SI4708_REG_CHANNEL,
		us,
		SI4708_REG_CHANNEL_CHANNLE_MASK
	);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_modify_reg() error\n",__FUNCTION__);
		return(rv);
	}

	return(0);
}

static int bcr_set_tune(UNIT_PARAMETERS * unit, int value)
{
	unsigned short us;
	int rv;

	print_trace("%s(%d)\n",__FUNCTION__,value);

	if(value)
		us = SI4708_REG_CHANNEL_TUNE;
	else
		us = 0;

	rv=rar_modify_reg(
		unit,
		SI4708_REG_CHANNEL,
		us,
		SI4708_REG_CHANNEL_TUNE
	);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_modify_reg() error\n",__FUNCTION__);
		return(rv);
	}

	if(value)
		msleep(SI4708_TUNE_COMP_TIME);
	else
		msleep(SI4708_TUNE_STOP_TIME);

	return(0);
}

static int bcr_wait_tune_complete(UNIT_PARAMETERS * unit)
{
	unsigned short us;
	int rv;
	int n;

	print_trace("%s()\n",__FUNCTION__);

	for(n=0; n<SI4708_TUNE_POLLING_COUNT; n++){

		rv=rar_read_reg(unit,SI4708_REG_STATUS_RSSI,&us);
		if(rv<0){
			printk(KERN_ERR "%s(): rar_read_reg() error\n",__FUNCTION__);
			return(rv);
		}

		if(us & SI4708_REG_STATUS_RSSI_STC)
			break;

		msleep(SI4708_STC_POLL_INTERVAL);
	}

	if(n==SI4708_TUNE_POLLING_COUNT)
		return(0);

	return(1);
}

static int bcr_check_signal_strength(UNIT_PARAMETERS * unit)
{
	unsigned short threshold;
	unsigned short now;
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	rv=rar_read_reg(
		unit,
		SI4708_REG_SYSCONFIG2,
		&threshold
	);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_read_reg() error\n",__FUNCTION__);
		return(rv);
	}

	threshold = (threshold & SI4708_REG_SYSCONFIG2_SEEKTH_MASK) >> 8;

	rv=rar_read_reg(
		unit,
		SI4708_REG_STATUS_RSSI,
		&now
	);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_read_reg() error\n",__FUNCTION__);
		return(rv);
	}

	if(now & SI4708_REG_STATUS_RSSI_AFCRL){
		/* signal invalid */
		return(0);
	}

	now = now & SI4708_REG_STATUS_RSSI_RSSI_MASK;

	return( (now>threshold) ? 1 : 0 );
}

#ifdef RSSI_RAW_VALUE_CONTRL
static int bcr_get_signal_strength(UNIT_PARAMETERS * unit, int * value)
{
	unsigned short us;
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	rv=rar_read_reg(
		unit,
		SI4708_REG_STATUS_RSSI,
		&us
	);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_read_reg() error\n",__FUNCTION__);
		return(rv);
	}

	if(us & SI4708_REG_STATUS_RSSI_AFCRL){
		/* signal invalid */
		*value=0;
		return(0);
	}

	*value = us & SI4708_REG_STATUS_RSSI_RSSI_MASK;

	return(0);
}
#endif /* RSSI_RAW_VALUE_CONTRL */
static int bcr_set_deemphasis_mode(UNIT_PARAMETERS * unit, int value)
{
	unsigned short us;
	int rv;

	print_trace("%s()\n",__FUNCTION__);

	if(value==DEEMPHASIS_MODE_50US)
		us=SI4708_REG_SYSCONFIG1_DE;
	else
		us=0;

	rv=rar_modify_reg(
		unit,
		SI4708_REG_SYSCONFIG1,
		us,
		SI4708_REG_SYSCONFIG1_DE
	);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_modify_reg() error\n",__FUNCTION__);
		return(rv);
	}

	return(0);
}

static int bcr_set_sense_mode(UNIT_PARAMETERS * unit, int value)
{
	unsigned short us;
	int rv;

	print_trace("%s(%d)\n",__FUNCTION__,value);

	if(value==SENSE_MODE_HIGH)
		us=model_parameters.sense_high<<8;
	else
		us=model_parameters.sense_low<<8;

	rv=rar_modify_reg(
		unit,
		SI4708_REG_SYSCONFIG2,
		us,
		SI4708_REG_SYSCONFIG2_SEEKTH_MASK
	);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_modify_reg() error\n",__FUNCTION__);
		return(rv);
	}

	return(0);
}

static int bcr_set_audio_mode(UNIT_PARAMETERS * unit, int value)
{
	unsigned short us;
	int rv;

	print_trace("%s(%d)\n",__FUNCTION__,value);

	if(value==AUDIO_MODE_MONO)
		us=SI4708_REG_POWERCFG_MONO;
	else
		us=0;

	rv=rar_modify_reg(
		unit,
		SI4708_REG_POWERCFG,
		us,
		SI4708_REG_POWERCFG_MONO
	);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_modify_reg() error\n",__FUNCTION__);
		return(rv);
	}

	msleep(SI4708_AUDIO_MODE_CHANGE_TIME);

	return(0);
}

static int bcr_switch_mute(UNIT_PARAMETERS * unit, int value)
{
	unsigned short us;
	int rv;

	print_trace("%s(%d)\n",__FUNCTION__,value);

	if(value==OFF)
		us=SI4708_REG_POWERCFG_DMUTE;
	else
		us=0;

	rv=rar_modify_reg(
		unit,
		SI4708_REG_POWERCFG,
		us,
		SI4708_REG_POWERCFG_DMUTE
	);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_modify_reg() error\n",__FUNCTION__);
		return(rv);
	}

	return(0);
}

static int bcr_set_power_mode(UNIT_PARAMETERS * unit, int value)
{
	unsigned short us;
	int rv;

	print_trace("%s(%d)\n",__FUNCTION__,value);

	if(value==POWER_MODE_ACTIVE)
		us=SI4708_REG_POWERCFG_ENABLE;
	else
		us=SI4708_REG_POWERCFG_ENABLE|SI4708_REG_POWERCFG_DISABLE;

	rv=rar_modify_reg(
		unit,
		SI4708_REG_POWERCFG,
		us,
		SI4708_REG_POWERCFG_ENABLE|SI4708_REG_POWERCFG_DISABLE
	);
	if(rv<0){
		printk(KERN_ERR "%s(): rar_modify_reg() error\n",__FUNCTION__);
		return(rv);
	}

	if(value==POWER_MODE_ACTIVE)
		msleep(SI4708_ENABLE_COMP_TIME);

	return(0);

}

static int bcr_reset_chip(UNIT_PARAMETERS * unit)
{
	SI4708ICX_PLATFORM_RESOURCE * resource;

	print_trace("%s()\n",__FUNCTION__);

	resource = unit->i2c_client->dev.platform_data;

#ifdef USE_MTK_GPIO_CONTROL
    mt_set_gpio_out(resource->port_reset,0);
#else
	gpio_set_value(resource->port_reset,0);
#endif

	msleep(SI4708_RESET_LOW_TIME);

#ifdef USE_MTK_GPIO_CONTROL
    mt_set_gpio_out(resource->port_reset,1);
#else
	gpio_set_value(resource->port_reset,1);
#endif

	return(0);
}

/*****************************/
/*@ register_access_routines */
/*****************************/

static int rar_modify_reg(
	UNIT_PARAMETERS * unit,
	unsigned char     address,
	unsigned short    value,
	unsigned short    mask
)
{
	unsigned short bak;
	unsigned short tmp;
	unsigned int   addr;
	int            rv;

	/* print_trace("%s()\n",__FUNCTION__); */

	addr=((unsigned int)address+SI4708_REG_COUNT-SI4708_REG_START)&0xF;

	rv=rar_read_all_reg(unit);
	if(rv<0) {
		printk(KERN_ERR "%s(): rar_read_all_reg() error\n",__FUNCTION__);
		return(rv);
	}

	bak=tmp=__be16_to_cpu(unit->reg_cache[addr]);

	tmp=(tmp&~mask)|(value&mask);

	unit->reg_cache[addr]=__cpu_to_be16(tmp);

	rv=rar_write_all_reg(unit);
	if(rv<0) {
		printk(KERN_ERR "%s(): rar_write_all_reg() error\n",__FUNCTION__);
		return(rv);
	}

	print_value(
		"%s: M %02X <- %04X <- %04X\n",
		MODULE_NAME,
		(unsigned int)address,
		(unsigned int)tmp,
		(unsigned int)bak
	);

	return(0);
}

static int rar_write_reg(
	UNIT_PARAMETERS * unit,
	unsigned char     address,
	unsigned short    value
)
{
	unsigned int addr;
	int rv;

	/* print_trace("%s()\n",__FUNCTION__); */

	addr=((unsigned int)address+SI4708_REG_COUNT-SI4708_REG_START)&0xF;

	unit->reg_cache[addr]=__cpu_to_be16(value);

	rv=rar_write_all_reg(unit);
	if(rv<0) {
		printk(KERN_ERR "%s(): rar_write_all_reg() error\n",__FUNCTION__);
		return(rv);
	}

	print_value(
		"%s: W %02X <- %04X\n",
		MODULE_NAME,
		(unsigned int)address,
		(unsigned int)value
	);

	return(0);
}

static int rar_read_reg(
	UNIT_PARAMETERS * unit,
	unsigned char     address,
	unsigned short  * value
)
{
	unsigned int addr;
	int rv;

	/* print_trace("%s()\n",__FUNCTION__); */

	addr=((unsigned int)address+SI4708_REG_COUNT-SI4708_REG_START)&0xF;

	rv=rar_read_all_reg(unit);
	if(rv<0) {
		printk(KERN_ERR "%s(): rar_read_all_reg() error\n",__FUNCTION__);
		return(rv);
	}

	*value=__be16_to_cpu(unit->reg_cache[addr]);

	print_value(
		"%s: R %02X -> %04X\n",
		MODULE_NAME,
		(unsigned int)address,
		(unsigned int)*value
	);

	return(0);
}

static int rar_write_all_reg(UNIT_PARAMETERS * unit)
{
	SI4708ICX_PLATFORM_RESOURCE * resource;
	struct i2c_client * i2c;
	struct i2c_msg      msg;
	int                 rv;

	/* print_trace("%s()\n",__FUNCTION__); */

	i2c=unit->i2c_client;
	resource=i2c->dev.platform_data;

	msg.addr  = i2c->addr;
	msg.flags = 0;
#ifdef USE_DMA
    msg.ext_flag = I2C_DMA_FLAG;
    msg.timing   = resource->i2c_timing;
#endif /* USE_DMA */
	msg.len   = sizeof(unsigned short)*SI4708_WREG_COUNT;
	msg.buf   = (unsigned char *)(unit->reg_cache+SI4708_WREG_OFFSET);

	rv=i2c_transfer(i2c->adapter,&msg,1);
	if(rv<0) {
		printk(KERN_ERR "%s(): i2c_transfer() error, rv = %d\n",__FUNCTION__,rv);
		return(rv);
	}

	if(rv!=1){
		printk(KERN_ERR "%s(): count mismacth, rv = %d\n",__FUNCTION__,rv);
		return(-1);
	}

	if(	( (unit->reg_cache[0x0] == 0xffff) && (unit->reg_cache[0x1] == 0xffff) ) || /* if two registers from front are 0xffff */
			( (unit->reg_cache[0xE] == 0xffff) && (unit->reg_cache[0xF] == 0xffff) ) || /* if two registers from last are 0xffff */
			( (unit->reg_cache[0xA] & 0x00ff) == 0x00ff ) /* if received signal strength is 0xff(normally indicates under 75) */
			){
		printk(KERN_ERR "%s(): restored register value is collapsed in si4708 driver!\n"
				"reset si4708 chip...", __FUNCTION__);
		reset_chip(unit);
	}

	return(0);
}

static int rar_read_all_reg(UNIT_PARAMETERS * unit)
{
	SI4708ICX_PLATFORM_RESOURCE * resource;
	struct i2c_client * i2c;
	struct i2c_msg      msg;
	int                 rv;

	/* print_trace("%s()\n",__FUNCTION__); */

	i2c=unit->i2c_client;
	resource=i2c->dev.platform_data;

	msg.addr  = i2c->addr;
#ifdef USE_DMA
    msg.ext_flag = I2C_DMA_FLAG;
    msg.timing   = resource->i2c_timing;
#endif /* USE_DMA */
	msg.flags = I2C_M_RD;
	msg.len   = sizeof(unsigned short)*SI4708_RREG_COUNT;
	msg.buf   = (unsigned char *)(unit->reg_cache+SI4708_RREG_OFFSET);

	rv=i2c_transfer(i2c->adapter,&msg,1);
	if(rv<0) {
		printk(KERN_ERR "%s(): i2c_transfer() error, rv = %d\n",__FUNCTION__,rv);
		return(rv);
	}

	if(rv!=1){
		printk(KERN_ERR "%s(): count mismacth, rv = %d\n",__FUNCTION__,rv);
		return(-1);
	}

	/* print_value("%04X %04X %04X %04X\n",unit->reg_cache[0x0],unit->reg_cache[0x1],unit->reg_cache[0x2],unit->reg_cache[0x3]); */
	/* print_value("%04X %04X %04X %04X\n",unit->reg_cache[0x4],unit->reg_cache[0x5],unit->reg_cache[0x6],unit->reg_cache[0x7]); */
	/* print_value("%04X %04X %04X %04X\n",unit->reg_cache[0x8],unit->reg_cache[0x9],unit->reg_cache[0xA],unit->reg_cache[0xB]); */
	/* print_value("%04X %04X %04X %04X\n",unit->reg_cache[0xC],unit->reg_cache[0xD],unit->reg_cache[0xE],unit->reg_cache[0xF]); */

	if(	( (unit->reg_cache[0x0] == 0xffff) && (unit->reg_cache[0x1] == 0xffff) ) || /* if two registers from front are 0xffff */
			( (unit->reg_cache[0xE] == 0xffff) && (unit->reg_cache[0xF] == 0xffff) ) || /* if two registers from last are 0xffff */
			( (unit->reg_cache[0xA] & 0x00ff) == 0x00ff ) /* if received signal strength is 0xff(normally indicates under 75) */
			){
		printk(KERN_ERR "%s(): i2c read value is collapsed in si4708 driver!\n", __FUNCTION__);
	}

	return(0);
}

/********************/
/*@ regmon_routines */
/********************/

#ifdef CONFIG_REGMON_DEBUG

static int si4708icx_regmon_write_reg(
	void         * private_data,
	unsigned int   address,
	unsigned int   value
);

static int si4708icx_regmon_read_reg(
	void         * private_data,
	unsigned int   address,
	unsigned int * value
);

static regmon_reg_info_t si4708icx_reg_info[] =
{
	{ "DEVICEID",    0x00 },
	{ "CHIPID",      0x01 },
	{ "POWERCFG",    0x02 },
	{ "CHANNEL",     0x03 },
	{ "SYSCONFIG1",  0x04 },
	{ "SYSCONFIG2",  0x05 },
	{ "SYSCONFIG3",  0x06 },
	{ "TEST1",       0x07 },
	{ "TEST2",       0x08 },
	{ "BOOTCONFIG",  0x09 },
	{ "STATUS_RSSI", 0x0A },
	{ "READCHAN",    0x0B },
	{ "RDSA",        0x0C },
	{ "RDSB",        0x0D },
	{ "RDSC",        0x0E },
	{ "RDSD",        0x0F },
};

static regmon_customer_info_t si4708icx_customer_info =
{
	.name           = SI4708ICX_RESOURCE_NAME,
	.reg_info       = si4708icx_reg_info,
	.reg_info_count = sizeof(si4708icx_reg_info)/sizeof(regmon_reg_info_t),
	.write_reg      = si4708icx_regmon_write_reg,
	.read_reg       = si4708icx_regmon_read_reg,
	.private_data   = NULL,
};

static int si4708icx_regmon_add(void * data)
{
	si4708icx_customer_info.private_data=data;
	regmon_add(&si4708icx_customer_info);
	return(0);
}

static int si4708icx_regmon_del(void)
{
	regmon_del(&si4708icx_customer_info);
	return(0);
}

static int si4708icx_regmon_write_reg(
	void         * private_data,
	unsigned int   address,
	unsigned int   value
)
{
	int rv;

	rv=rar_write_reg(
		(UNIT_PARAMETERS *)private_data,
		(unsigned char)address,
		(unsigned short)value
	);

	return(rv);
}

static int si4708icx_regmon_read_reg(
	void         * private_data,
	unsigned int   address,
	unsigned int * value
)
{
	unsigned short us;
	int rv;

	rv=rar_read_reg(
		(UNIT_PARAMETERS *)private_data,
		(unsigned char)address,
		&us
	);

	*value=us;

	return(rv);
}

#endif
